from functools import reduce

def split(curve, s, t, n=None, I=None):
  """
  Perform Shamir sharing on the secret `s` to the degree `t - 1` split `n`
  ways. The split secret requires `t` shares to be reconstructed.

  :param s: Secret to split.
  :type s: int
  :param t: Share threshold required to reconstruct secret.
  :type t: int
  :param n: Total number of shares to split secret into.
  :type n: int, optional
  :param I: Indices used for each party. Defaults to range(1, 1 + n).
  :type I: list, optional
  :return: Dictionary of shares. Each key is an int in the range 1<=x<=n
    representing that share's free term.
  :rtype dict:
  """
  if I == None:
    assert n != None
    I = range(1, 1 + n)
  elif n == None:
    n = len(I)
  assert t > 1
  assert t <= n
  coefs = [curve.scalar_random() for i in range(0, t - 1)] + [s]
  shares = dict()
  for x in I:
    shares[x] = reduce(
      lambda partial, coef: \
      curve.scalar_add(coef, curve.scalar_mul(partial, x)),
      coefs,
    )
  return shares

def combine(curve, shares):
  """
  Reconstitute a secret from a dictionary of shares. The number of shares must
  be equal to `t` to reconstitute the original secret.

  :param shares: Dictionary of shares. Each key is the free term of the share.
  :type shares: dict
  :return: Reconstituted secret.
  :rtype: int
  """
  s = 0
  for xi in shares:
    yi = shares[xi]
    num = 1
    for xj in shares:
      if xi != xj:
        num = curve.scalar_mul(num, xj)
    denum = 1
    for xj in shares:
      if xi != xj:
        denum = curve.scalar_mul(denum, curve.scalar_sub(xj, xi))
    s = curve.scalar_add(
      curve.scalar_mul(curve.scalar_mul(num, curve.scalar_invert(denum)), yi),
      s,
    )
  return s



import argparse, base58, enum, json
from functools import partial

from . import Ecdsa, Eddsa, curves, serialization

class HelpFormatter(argparse.HelpFormatter):
  def add_argument(self, action):
    if action.help is not argparse.SUPPRESS:
      get_invocation = self._format_action_invocation
      invocations = [get_invocation(action)]
      current_indent = self._current_indent
      for subaction in self._iter_indented_subactions(action):
        indent_chg = self._current_indent - current_indent
        added_indent = 'x' * indent_chg
        invocations.append(added_indent+get_invocation(subaction))
      invocation_length = max([len(s) for s in invocations])
      action_length = invocation_length + self._current_indent
      self._action_max_length = max(self._action_max_length, action_length)
      self._add_item(self._format_action, [action])

def main():
  formatter = lambda prog: HelpFormatter(prog)

  parser = argparse.ArgumentParser(prog='ggmpc', formatter_class=formatter)

  command_subparsers = parser.add_subparsers(
    dest='COMMAND',
    required=True,
  )

  ecdsa_parser = command_subparsers.add_parser(
    'ecdsa',
    help='ECDSA commands',
  )
  ecdsa_parser.add_argument(
    '-c',
    help='specify curve',
    choices=['secp256k1'],
    default='secp256k1',
  )

  eddsa_parser = command_subparsers.add_parser(
    'eddsa',
    help='EdDSA commands',
  )
  eddsa_parser.add_argument(
    '-c',
    help='specify curve',
    choices=['ed25519'],
    default='ed25519',
  )

  # ECDSA subcommands.

  ecdsa_subparsers = ecdsa_parser.add_subparsers(
    dest='SUBCOMMAND',
    metavar='SUBCOMMAND',
    required=True,
  )

  keyshare_parser = ecdsa_subparsers.add_parser(
    'keyshare',
    help='create key shares',
  )
  keyshare_parser.set_defaults(func=ecdsa_key_share)
  keyshare_parser.add_argument(
    '-i',
    type=int,
    help='player index',
    required=True,
  )
  keyshare_parser.add_argument(
    '-t',
    type=int,
    help='required threshold',
    required=True,
  )
  keyshare_parser.add_argument(
    '-n',
    type=int,
    help='number of players',
    required=True,
  )
  keyshare_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  keycombine_parser = ecdsa_subparsers.add_parser(
    'keycombine',
    help='combine key shares',
  )
  keycombine_parser.set_defaults(func=ecdsa_key_combine)
  keycombine_parser.add_argument(
    'KEYSHARE',
    type=str,
    help='your key share and shares received from other players',
    nargs='+',
  )
  keycombine_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  signshare_parser = ecdsa_subparsers.add_parser(
    'signshare',
    help='create signing shares',
  )
  signshare_parser.set_defaults(func=ecdsa_sign_share)
  signshare_parser.add_argument(
    'KEYSHARE',
    type=str,
    help='combined key shares for all signers',
    nargs='+',
  )
  signshare_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  signconvert_parser = ecdsa_subparsers.add_parser(
    'signconvert',
    help='convert signing shares',
  )
  signconvert_parser.set_defaults(func=ecdsa_sign_convert)
  signconvert_parser.add_argument(
    'SIGNSHARE',
    type=str,
    help='your signing share and shares received from other signers',
    nargs='+',
  )
  signconvert_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  signcombine_parser = ecdsa_subparsers.add_parser(
    'signcombine',
    help='combine converted signing shares and signature shares',
  )
  signcombine_parser.set_defaults(func=ecdsa_sign_combine)
  signcombine_parser.add_argument(
    'SIGNSHARE',
    type=str,
    help='your signing shares and shares received from other signers',
    nargs='+',
  )
  signcombine_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  sign_parser = ecdsa_subparsers.add_parser(
    'sign',
    help='sign message using converted signing shares',
  )
  sign_parser.set_defaults(func=ecdsa_sign)
  sign_parser.add_argument(
    'MESSAGE',
    type=str,
    help='message to sign',
    nargs=1,
  )
  sign_parser.add_argument(
    'SIGNSHARE',
    type=str,
    help='your combined signing shares',
    nargs='+',
  )
  sign_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  verify_parser = ecdsa_subparsers.add_parser(
    'verify',
    help='verify a signature',
  )
  verify_parser.set_defaults(func=ecdsa_verify)
  verify_parser.add_argument(
    'MESSAGE',
    type=str,
    help='message to verify',
    nargs=1,
  )
  verify_parser.add_argument(
    'SIGNATURE',
    type=str,
    help='signature to verify',
    nargs=1,
  )

  # EdDSA subcommands.

  eddsa_subparsers = eddsa_parser.add_subparsers(
    dest='SUBCOMMAND',
    metavar='SUBCOMMAND',
    required=True,
  )

  keyshare_parser = eddsa_subparsers.add_parser(
    'keyshare',
    help='create key shares',
  )
  keyshare_parser.set_defaults(func=eddsa_key_share)
  keyshare_parser.add_argument(
    '-i',
    type=int,
    help='player index',
    required=True,
  )
  keyshare_parser.add_argument(
    '-t',
    type=int,
    help='required threshold',
    required=True,
  )
  keyshare_parser.add_argument(
    '-n',
    type=int,
    help='number of players',
    required=True,
  )
  keyshare_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  keycombine_parser = eddsa_subparsers.add_parser(
    'keycombine',
    help='combine key shares',
  )
  keycombine_parser.set_defaults(func=eddsa_key_combine)
  keycombine_parser.add_argument(
    'KEYSHARE',
    type=str,
    help='your key share and shares received from other players',
    nargs='+',
  )
  keycombine_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  signshare_parser = eddsa_subparsers.add_parser(
    'signshare',
    help='create signing shares',
  )
  signshare_parser.set_defaults(func=eddsa_sign_share)
  signshare_parser.add_argument(
    'MESSAGE',
    type=str,
    help='message to sign',
    nargs=1,
  )
  signshare_parser.add_argument(
    'KEYSHARE',
    type=str,
    help='key shares for all signers',
    nargs='+',
  )
  signshare_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  sign_parser = eddsa_subparsers.add_parser(
    'sign',
    help='sign message using signing shares',
  )
  sign_parser.set_defaults(func=eddsa_sign)
  sign_parser.add_argument(
    'MESSAGE',
    type=str,
    help='message to sign',
    nargs=1,
  )
  sign_parser.add_argument(
    'SIGNSHARE',
    type=str,
    help='signing shares',
    nargs='+',
  )
  sign_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  signcombine_parser = eddsa_subparsers.add_parser(
    'signcombine',
    help='combine signature shares',
  )
  signcombine_parser.set_defaults(func=eddsa_sign_combine)
  signcombine_parser.add_argument(
    'SIGNSHARE',
    type=str,
    help='your signing shares and shares received from other signers',
    nargs='+',
  )
  signcombine_parser.add_argument(
    '-nz',
    help='disable compression',
    dest='compress',
    action='store_false',
    default=True,
  )

  verify_parser = eddsa_subparsers.add_parser(
    'verify',
    help='verify a signature',
  )
  verify_parser.set_defaults(func=eddsa_verify)
  verify_parser.add_argument(
    'MESSAGE',
    type=str,
    help='message to verify',
    nargs=1,
  )
  verify_parser.add_argument(
    'SIGNATURE',
    type=str,
    help='signature to verify',
    nargs=1,
  )

  # Deserialize command

  deserialize_parser = command_subparsers.add_parser(
    'deserialize',
    help='deserialize data',
  )
  deserialize_parser.add_argument(
    'DATA',
    type=str,
    help='data to deserialize',
    nargs=1,
  )

  args = parser.parse_args()

  if args.COMMAND == 'ecdsa':
    args.func(Ecdsa(getattr(curves, args.c)), args)
  elif args.COMMAND == 'eddsa':
    args.func(Eddsa(getattr(curves, args.c)), args)
  elif args.COMMAND == 'deserialize':
    command_deserialize(args)
  
def ecdsa_key_share(mpc, args):
  ser = partial(serialize, compress=args.compress)
  shares = mpc.key_share(args.i, args.t, args.n)
  P_i = shares[args.i]
  print('\nSave for yourself: %s' % ser(DataType.ECDSA_P_SHARE, P_i))
  for i in shares:
    if i != args.i:
      print(
        '\nSend to player %d: %s' \
        % (i, ser(DataType.ECDSA_N_SHARE, shares[i]))
      )

def ecdsa_key_combine(mpc, args):
  ser = partial(serialize, compress=args.compress)
  shares = mpc.key_combine(list(map(deserialize, args.KEYSHARE)))
  P_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
  print('\nSave for yourself: %s' % ser(DataType.ECDSA_X_SHARE, P_i))
  for i in shares:
    if i != P_i['i']:
      print(
        '\nSave for player %d: %s' \
        % (i, ser(DataType.ECDSA_Y_SHARE, shares[i]))
      )

def ecdsa_sign_share(mpc, args):
  ser = partial(serialize, compress=args.compress)
  shares = mpc.sign_share(list(map(deserialize, args.KEYSHARE)))
  P_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
  print('\nSave for yourself: %s' % ser(DataType.ECDSA_W_SHARE, P_i))
  for i in shares:
    if i != P_i['i']:
      print(
        '\nSend to player %d: %s' \
        % (i, ser(DataType.ECDSA_K_SHARE, shares[i]))
      )

def ecdsa_sign_convert(mpc, args):
  ser = partial(serialize, compress=args.compress)
  shares = list(map(deserialize, args.SIGNSHARE))
  for i in range(0, len(shares)):
    share = shares[i]
    if type(list(share.keys())[0]) == int:
      shares[i] = shares[i][next(filter(lambda i: not 'j' in share[i], share))]
  shares = mpc.sign_convert(shares)
  P_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
  P_j = shares[next(filter(lambda i: 'j' in shares[i], shares))]
  data_type = None
  if 'alpha' in P_i:
    data_type = DataType.ECDSA_G_SHARE
  else:
    data_type = DataType.ECDSA_B_SHARE
  print('\nSave for yourself:', ser(data_type, shares),)
  data_type = None
  if 'k' in P_j or 'alpha' in P_j:
    if 'k' in P_j:
      data_type = DataType.ECDSA_A_SHARE
    else:
      data_type = DataType.ECDSA_M_SHARE
    print('\nSend to player %d: %s' % (P_j['i'], ser(data_type, P_j)))

def ecdsa_sign_combine(mpc, args):
  ser = partial(serialize, compress=args.compress)
  shares = list(map(deserialize, args.SIGNSHARE))
  if 'r' in shares[0] and 's' in shares[0]:
    print('\n%s' % ser(DataType.ECDSA_SIGNATURE, mpc.sign_combine(shares)))
  else:
    shares = mpc.sign_combine(shares)
    P_i = shares[next(filter(lambda i: 'k' in shares[i], shares))]
    print('\nSave for yourself: %s' % ser(DataType.ECDSA_O_SHARE, P_i))
    for i in shares:
      if i != P_i['i']:
        print(
          '\nSend to player %d: %s' \
          % (i, ser(DataType.ECDSA_D_SHARE, shares[i]))
        )
                                            
def ecdsa_sign(mpc, args):
  ser = partial(serialize, compress=args.compress)
  M = args.MESSAGE[0].encode('ascii')
  shares = list(map(deserialize, args.SIGNSHARE))
  print(
    '\nSend to all players: %s' \
    % ser(DataType.ECDSA_S_SHARE, mpc.sign(M, shares))
  )

def ecdsa_verify(mpc, args):
  M = args.MESSAGE[0].encode('ascii')
  assert mpc.verify(M, deserialize(args.SIGNATURE[0]))
  print('\nVerification succeeded')

def eddsa_key_share(mpc, args):
  ser = partial(serialize, compress=args.compress)
  shares = mpc.key_share(args.i, args.t, args.n)
  P_i = shares[args.i]
  print('\nSave for yourself: %s' % ser(DataType.EDDSA_U_SHARE, P_i))
  for i in shares:
    if i != args.i:
      print(
        '\nSend to player %d: %s' \
        % (i, ser(DataType.EDDSA_Y_SHARE, shares[i]))
      )

def eddsa_key_combine(mpc, args):
  ser = partial(serialize, compress=args.compress)
  shares = mpc.key_combine(list(map(deserialize, args.KEYSHARE)))
  P_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
  print('\nSave for yourself: %s' % ser(DataType.EDDSA_P_SHARE, P_i))
  for i in shares:
    if i != P_i['i']:
      print(
        '\nSave for player %d: %s' \
        % (i, ser(DataType.EDDSA_J_SHARE, shares[i]))
      )

def eddsa_sign_share(mpc, args):
  ser = partial(serialize, compress=args.compress)
  M = args.MESSAGE[0].encode('ascii')
  shares = mpc.sign_share(M, list(map(deserialize, args.KEYSHARE)))
  P_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
  print('\nSave for yourself: %s' % ser(DataType.EDDSA_X_SHARE, P_i))
  for i in shares:
    if i != P_i['i']:
      print(
        '\nSend to player %d: %s' \
        % (i, ser(DataType.EDDSA_R_SHARE, shares[i]))
      )

def eddsa_sign(mpc, args):
  ser = partial(serialize, compress=args.compress)
  M = args.MESSAGE[0].encode('ascii')
  shares = list(map(deserialize, args.SIGNSHARE))
  print(
    '\nSend to all players: %s' \
    % ser(DataType.EDDSA_G_SHARE, mpc.sign(M, shares))
  )

def eddsa_sign_combine(mpc, args):
  ser = partial(serialize, compress=args.compress)
  shares = list(map(deserialize, args.SIGNSHARE))
  print(
    '\n%s' \
    % ser(DataType.EDDSA_SIGNATURE, mpc.sign_combine(shares))
  )

def eddsa_verify(mpc, args):
  M = args.MESSAGE[0].encode('ascii')
  assert mpc.verify(M, deserialize(args.SIGNATURE[0]))
  print('\nVerification succeeded')

def command_deserialize(args):
  print('\n%s' % json.dumps(encode(deserialize(args.DATA[0])), indent=2))

def encode(obj):
  encoded = {}
  for key in obj:
    val = obj[key]
    if type(val) == dict:
      encoded[key] = encode(val)
    elif key == 'i':
      encoded[key] = val
    elif key == 'j':
      encoded[key] = val
    elif key == 'y' or key == 'Gamma':
      if type(val) == int:
        encoded[key] = serialization.serialize_int(val, 32).hex()
      else:
        encoded[key] = serialization.ecdsa.serialize_point(val).hex()
    else:
      length = (val.bit_length() + 7) // 8
      encoded[key] = serialization.serialize_int(val, length).hex()
  return encoded

def serialize_to_json(obj):
  return '\'' + json.dumps(encode(obj), separators=(',',':')) + '\''

def decode(encoded):
  obj = {}
  for key in encoded:
    val = encoded[key]
    if key.isnumeric():
      key = int(key)
    if type(val) == dict:
      obj[key] = decode(val)
    elif key == 'i':
      obj[key] = val
    elif key == 'j':
      obj[key] = val
    elif key == 'y' or key == 'Gamma':
      ser = bytes.fromhex(val)
      if len(ser) == 33:
        obj[key] = serialization.ecdsa.deserialize_point(ser)[1]
      else:
        obj[key] = serialization.deserialize_int(ser, 32)[1]
    else:
      ser = bytes.fromhex(val)
      obj[key] = serialization.deserialize_int(ser, len(ser))[1]
  return obj

def deserialize_from_json(string):
  return decode(json.loads(string))

class DataType(enum.Enum):
  ECDSA_P_SHARE = 'pshc', '_i0_'
  ECDSA_N_SHARE = 'nshc', '_i0j1_'
  ECDSA_X_SHARE = 'xshc', '_i0_'
  ECDSA_Y_SHARE = 'yshc', '_i0j1_'
  ECDSA_W_SHARE = 'wshc', '_i0_'
  ECDSA_K_SHARE = 'kshc', '_i0j1_'
  ECDSA_B_SHARE = 'bshc', '_i0j1_'
  ECDSA_A_SHARE = 'ashc', '_i0j1_'
  ECDSA_M_SHARE = 'mshc', '_i0j1_'
  ECDSA_G_SHARE = 'gshc', '_i0j1_'
  ECDSA_O_SHARE = 'oshc', '_i0_'
  ECDSA_D_SHARE = 'dshc', '_i0j1_'
  ECDSA_S_SHARE = 'sshc', '_i0_'
  ECDSA_SIGNATURE = 'sigc_',
  EDDSA_U_SHARE = 'ushd', '_i0_'
  EDDSA_Y_SHARE = 'yshd', '_i0j1_'
  EDDSA_P_SHARE = 'pshd', '_i0_'
  EDDSA_J_SHARE = 'jshd', '_i0j1'
  EDDSA_X_SHARE = 'xshd', '_i0_'
  EDDSA_R_SHARE = 'rshd', '_i0j1_'
  EDDSA_G_SHARE = 'gshd', '_i0_'
  EDDSA_SIGNATURE = 'sigd_',

alphabet = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'

def serialize(data_type, data, compress=True):
  if compress:
    serialize = {
      DataType.ECDSA_P_SHARE: serialization.ecdsa.serialize_p_share,
      DataType.ECDSA_N_SHARE: serialization.ecdsa.serialize_n_share,
      DataType.ECDSA_X_SHARE: serialization.ecdsa.serialize_x_share,
      DataType.ECDSA_Y_SHARE: serialization.ecdsa.serialize_y_share,
      DataType.ECDSA_W_SHARE: serialization.ecdsa.serialize_w_share,
      DataType.ECDSA_K_SHARE: serialization.ecdsa.serialize_k_share,
      DataType.ECDSA_B_SHARE: serialization.ecdsa.serialize_b_share,
      DataType.ECDSA_A_SHARE: serialization.ecdsa.serialize_a_share,
      DataType.ECDSA_M_SHARE: serialization.ecdsa.serialize_m_share,
      DataType.ECDSA_G_SHARE: serialization.ecdsa.serialize_g_share,
      DataType.ECDSA_O_SHARE: serialization.ecdsa.serialize_o_share,
      DataType.ECDSA_D_SHARE: serialization.ecdsa.serialize_d_share,
      DataType.ECDSA_S_SHARE: serialization.ecdsa.serialize_s_share,
      DataType.ECDSA_SIGNATURE: serialization.ecdsa.serialize_signature,
      DataType.EDDSA_U_SHARE: serialization.eddsa.serialize_u_share,
      DataType.EDDSA_Y_SHARE: serialization.eddsa.serialize_y_share,
      DataType.EDDSA_P_SHARE: serialization.eddsa.serialize_p_share,
      DataType.EDDSA_J_SHARE: serialization.eddsa.serialize_j_share,
      DataType.EDDSA_X_SHARE: serialization.eddsa.serialize_x_share,
      DataType.EDDSA_R_SHARE: serialization.eddsa.serialize_r_share,
      DataType.EDDSA_G_SHARE: serialization.eddsa.serialize_g_share,
      DataType.EDDSA_SIGNATURE: serialization.eddsa.serialize_signature,
    }
    data = serialize[data_type](data)
    i, j = None, None
    if len(data_type.value) > 1:
      if data_type.value[1].find('1') >= 0:
        j = data[1]
        data = data[:1] + data[2:]
      if data_type.value[1].find('0') >= 0:
        i = data[0]
        data = data[1:]
    prefix = data_type.value[0]
    if len(data_type.value) > 1:
      prefix += data_type.value[1]
      if i != None:
        index = len(data_type.value[0]) + data_type.value[1].find('0')
        prefix = prefix[:index] + alphabet[i - 1] + prefix[index + 1:]
      if j != None:
        index = len(data_type.value[0]) + data_type.value[1].find('1')
        prefix = prefix[:index] + alphabet[j - 1] + prefix[index + 1:]
    if len(data) > 0:
      return prefix + base58.b58encode_check(data).decode('ascii')
    return prefix
  return serialize_to_json(data)

def deserialize(data):
  if data[0] == '{':
    return deserialize_from_json(data)
  deserialize = {
    DataType.ECDSA_P_SHARE: serialization.ecdsa.deserialize_p_share,
    DataType.ECDSA_N_SHARE: serialization.ecdsa.deserialize_n_share,
    DataType.ECDSA_X_SHARE: serialization.ecdsa.deserialize_x_share,
    DataType.ECDSA_Y_SHARE: serialization.ecdsa.deserialize_y_share,
    DataType.ECDSA_W_SHARE: serialization.ecdsa.deserialize_w_share,
    DataType.ECDSA_K_SHARE: serialization.ecdsa.deserialize_k_share,
    DataType.ECDSA_B_SHARE: serialization.ecdsa.deserialize_b_share,
    DataType.ECDSA_A_SHARE: serialization.ecdsa.deserialize_a_share,
    DataType.ECDSA_M_SHARE: serialization.ecdsa.deserialize_m_share,
    DataType.ECDSA_G_SHARE: serialization.ecdsa.deserialize_g_share,
    DataType.ECDSA_O_SHARE: serialization.ecdsa.deserialize_o_share,
    DataType.ECDSA_D_SHARE: serialization.ecdsa.deserialize_d_share,
    DataType.ECDSA_S_SHARE: serialization.ecdsa.deserialize_s_share,
    DataType.ECDSA_SIGNATURE: serialization.ecdsa.deserialize_signature,
    DataType.EDDSA_U_SHARE: serialization.eddsa.deserialize_u_share,
    DataType.EDDSA_Y_SHARE: serialization.eddsa.deserialize_y_share,
    DataType.EDDSA_P_SHARE: serialization.eddsa.deserialize_p_share,
    DataType.EDDSA_J_SHARE: serialization.eddsa.deserialize_j_share,
    DataType.EDDSA_X_SHARE: serialization.eddsa.deserialize_x_share,
    DataType.EDDSA_R_SHARE: serialization.eddsa.deserialize_r_share,
    DataType.EDDSA_G_SHARE: serialization.eddsa.deserialize_g_share,
    DataType.EDDSA_SIGNATURE: serialization.eddsa.deserialize_signature,
  }
  data_size = {
    DataType.ECDSA_P_SHARE: 450,
    DataType.ECDSA_N_SHARE: 451,
    DataType.ECDSA_X_SHARE: 450,
    DataType.ECDSA_Y_SHARE: 386,
    DataType.ECDSA_W_SHARE: 514,
    DataType.ECDSA_K_SHARE: 1154,
    DataType.ECDSA_B_SHARE: 3267,
    DataType.ECDSA_A_SHARE: 2690,
    DataType.ECDSA_M_SHARE: 1538,
    DataType.ECDSA_G_SHARE: 259,
    DataType.ECDSA_O_SHARE: 163,
    DataType.ECDSA_D_SHARE: 67,
    DataType.ECDSA_S_SHARE: 98,
    DataType.ECDSA_SIGNATURE: 97,
    DataType.EDDSA_U_SHARE: 97,
    DataType.EDDSA_Y_SHARE: 66,
    DataType.EDDSA_P_SHARE: 97,
    DataType.EDDSA_J_SHARE: 2,
    DataType.EDDSA_X_SHARE: 129,
    DataType.EDDSA_R_SHARE: 66,
    DataType.EDDSA_G_SHARE: 97,
    DataType.EDDSA_SIGNATURE: 96,
  }
  for data_type in DataType:
    if data[:len(data_type.value[0])] == data_type.value[0]:
      prefix_len = len(data_type.value[0])
      prefix = data[:prefix_len]
      i, j = None, None
      index_len = 0
      if len(data_type.value) > 1:
        prefix += data[prefix_len:prefix_len + len(data_type.value[1])]
        prefix_len += len(data_type.value[1])
        index = prefix[len(data_type.value[0]):]
        if data_type.value[1].find('0') >= 0:
          i = alphabet.index(index[data_type.value[1].find('0')]) + 1
        if data_type.value[1].find('1') >= 0:
          j = alphabet.index(index[data_type.value[1].find('1')]) + 1
      if len(data) > prefix_len:
        data = base58.b58decode_check(data[prefix_len:])
      else:
        data = b''
      if j != None:
        data = j.to_bytes(1, 'big') + data
      if i != None:
        data = i.to_bytes(1, 'big') + data
      rem, data = deserialize[data_type](data)
      assert len(rem) == 0
      return data

if __name__ == '__main__':
  main()



""" Supported elliptic curves. """

import ecdsa, hashlib, random
from nacl import bindings

class Curve:
  """
  Elliptic curve parameters.

  :param scalar_random: Function that returns a random field element.

  :type scalar_random: function

  :param scalar_add: Function that returns the sum of two field elements modulo
    the order of the curve.

  :type scalar_add: function

  :param scalar_sub: Function that returns the difference of two field elements
    modulo the order of the curve.

  :type scalar_sub: function

  :param scalar_mul: Function that returns the product of two field elements
    modulo the order of the curve.

  :type scalar_mul: function

  :param scalar_reduce: Function that reduces a scalar modulo the order of the
    curve.

  :type scalar_reduce: function

  :param scalar_invert: Function that returns the modular multiplicative inverse
    of a field element.

  :type scalar_invert: function

  :param scalar_negate: Function that returns the negated field element modulo
    the order of the curve.

  :type scalar_negate: function

  :param point_add: Function that adds two group elements.

  :type point_add: function

  :param point_mul: Function that multiplies a group element by a field element.

  :type point_mul: function

  :param point_mul_base: Function that multiplies the curve generator by a field
    element.

  :type point_mul_base: function

  :param order: Function that returns the order of the curve.

  :type order: function

  :param hash: A hash function using a `hashlib`-like interface.

  :type hash: function

  :param verify: Function that verifies a signature.

  :type verify: function
  """

  def __init__(
      self,
      scalar_random,
      scalar_add,
      scalar_sub,
      scalar_mul,
      scalar_reduce,
      scalar_negate,
      scalar_invert,
      point_add,
      point_mul,
      point_mul_base,
      order,
      hash,
      verify,
  ):
    self.scalar_random = scalar_random
    self.scalar_add = scalar_add
    self.scalar_sub = scalar_sub
    self.scalar_mul = scalar_mul
    self.scalar_reduce = scalar_reduce
    self.scalar_negate = scalar_negate
    self.scalar_invert = scalar_invert
    self.point_add = point_add
    self.point_mul = point_mul
    self.point_mul_base = point_mul_base
    self.order = order
    self.hash = hash
    self.verify = verify

def _ecdsa_verify(curve, hash):
  return lambda y, M, sig: ecdsa.VerifyingKey.from_public_point(
    y,
    curve=curve,
    hashfunc=hash,
  ).verify(sig, M)

_secp256k1_order = int(ecdsa.ecdsa.generator_secp256k1.order())

secp256k1 = Curve(
  lambda: random.SystemRandom().randrange(1, _secp256k1_order),
  lambda a, b: (a + b) % _secp256k1_order,
  lambda a, b: (a - b) % _secp256k1_order,
  lambda a, b: (a * b) % _secp256k1_order,
  lambda x: x % _secp256k1_order,
  lambda x: -x % _secp256k1_order,
  lambda x: pow(x, -1, _secp256k1_order),
  lambda p, q: p + q,
  lambda p, q: p * q,
  lambda n: ecdsa.ecdsa.generator_secp256k1 * n,
  lambda: _secp256k1_order,
  hashlib.sha256,
  _ecdsa_verify(ecdsa.curves.SECP256k1, hashlib.sha256),
)
""" Secp256k1 ECDSA curve. """

ed25519 = Curve(
  lambda: int.from_bytes(bindings.crypto_core_ed25519_scalar_reduce(
    random.SystemRandom().randrange(1, 2 ** 512).to_bytes(64, 'little'),
  ), 'little'),
  lambda a, b: int.from_bytes(bindings.crypto_core_ed25519_scalar_add(
    a.to_bytes(32, 'little'),
    b.to_bytes(32, 'little'),
  ), 'little'),
  lambda a, b: int.from_bytes(bindings.crypto_core_ed25519_scalar_sub(
    a.to_bytes(32, 'little'),
    b.to_bytes(32, 'little'),
  ), 'little'),
  lambda a, b: int.from_bytes(bindings.crypto_core_ed25519_scalar_mul(
    a.to_bytes(32, 'little'),
    b.to_bytes(32, 'little'),
  ), 'little'),
  lambda x: int.from_bytes(bindings.crypto_core_ed25519_scalar_reduce(
    x.to_bytes(64, 'little'),
  ), 'little'),
  lambda x: int.from_bytes(bindings.crypto_core_ed25519_scalar_negate(
    x.to_bytes(32, 'little'),
  ), 'little'),
  lambda x: int.from_bytes(bindings.crypto_core_ed25519_scalar_invert(
    x.to_bytes(32, 'little'),
  ), 'little'),
  lambda p, q: int.from_bytes(bindings.crypto_core_ed25519_add(
    p.to_bytes(32, 'little'),
    q.to_bytes(32, 'little'),
  ), 'little'),
  lambda p, n: int.from_bytes(bindings.crypto_scalarmult_ed25519_noclamp(
    p.to_bytes(32, 'little'),
    n.to_bytes(32, 'little'),
  ), 'little'),
  lambda n: int.from_bytes(bindings.crypto_scalarmult_ed25519_base_noclamp(
    n.to_bytes(32, 'little'),
  ), 'little'),
  None,
  hashlib.sha512,
  lambda y, M, sig: bindings.crypto_sign_open(sig + M, y)
)
""" Ed25519 EdDSA curve. """



import ecdsa, unittest
import src.ggmpc as ggmpc
import src.ggmpc.curves as curves

class Ecdsa(unittest.TestCase):

  def test_2_in_2_of_3(self):
    mpc = ggmpc.Ecdsa(curves.secp256k1)

    A = mpc.key_share(1, 2, 3)
    B = mpc.key_share(2, 2, 3)
    C = mpc.key_share(3, 2, 3)

    A, B, C = \
      mpc.key_combine((A[1], B[1], C[1])), \
      mpc.key_combine((A[2], B[2], C[2])), \
      mpc.key_combine((A[3], B[3], C[3])),

    AB = mpc.sign_share((A[1], A[2]))
    BA = mpc.sign_convert((B[1], B[2], AB[2]))
    AB = mpc.sign_convert((AB[1], BA[1]))
    BA = mpc.sign_convert((BA[2], AB[2]))

    AB, BA = mpc.sign_combine((AB,)), mpc.sign_combine((BA,))

    M = b'MPC on a Friday night'

    A, B = mpc.sign(M, (AB[1], BA[1])), mpc.sign(M, (AB[2], BA[2])),

    sig = mpc.sign_combine((A, B))

    assert mpc.verify(M, sig)

  def test_2_in_3_of_5(self):
    mpc = ggmpc.Ecdsa(curves.secp256k1)
    
    A = mpc.key_share(1, 3, 5)
    B = mpc.key_share(2, 3, 5)
    C = mpc.key_share(3, 3, 5)
    D = mpc.key_share(4, 3, 5)
    E = mpc.key_share(5, 3, 5)

    A, B, C, D, E = \
      mpc.key_combine((A[1], B[1], C[1], D[1], E[1])), \
      mpc.key_combine((A[2], B[2], C[2], D[2], E[2])), \
      mpc.key_combine((A[3], B[3], C[3], D[3], E[3])), \
      mpc.key_combine((A[4], B[4], C[4], D[4], E[4])), \
      mpc.key_combine((A[5], B[5], C[4], D[4], E[4])),

    AC = mpc.sign_share((A[1], A[2], A[3]))
    CA = mpc.sign_convert((C[1], C[2], C[3], AC[3]))
    AC = mpc.sign_convert((AC[1], CA[1]))
    CA = mpc.sign_convert((CA[3], AC[3]))

    AC, CA = mpc.sign_combine((AC,)), mpc.sign_combine((CA,))

    M = b'MPC on a Friday night'

    A, C = mpc.sign(M, (AC[1], CA[1])), mpc.sign(M, (AC[3], CA[3]))

    sig = mpc.sign_combine((A, C))

    try:
      assert not mpc.verify(M, sig)
    except ecdsa.keys.BadSignatureError:
      pass

  def test_3_in_3_of_5(self):
    mpc = ggmpc.Ecdsa(curves.secp256k1)
    
    A = mpc.key_share(1, 3, 5)
    B = mpc.key_share(2, 3, 5)
    C = mpc.key_share(3, 3, 5)
    D = mpc.key_share(4, 3, 5)
    E = mpc.key_share(5, 3, 5)

    A, B, C, D, E = \
      mpc.key_combine((A[1], B[1], C[1], D[1], E[1])), \
      mpc.key_combine((A[2], B[2], C[2], D[2], E[2])), \
      mpc.key_combine((A[3], B[3], C[3], D[3], E[3])), \
      mpc.key_combine((A[4], B[4], C[4], D[4], E[4])), \
      mpc.key_combine((A[5], B[5], C[4], D[4], E[4])),

    A = mpc.sign_share((A[1], A[2], A[3]))
    B = mpc.sign_share((B[1], B[2], B[3]))
    C = mpc.sign_share((C[1], C[2], C[3]))

    AB = mpc.sign_convert((A[1], B[1]))
    BA = mpc.sign_convert((B[2], AB[2]))
    AB = mpc.sign_convert((AB[1], BA[1]))

    AC = mpc.sign_convert((A[1], C[1]))
    CA = mpc.sign_convert((C[3], AC[3]))
    AC = mpc.sign_convert((AC[1], CA[1]))

    BC = mpc.sign_convert((B[2], C[2]))
    CB = mpc.sign_convert((C[3], BC[3]))
    BC = mpc.sign_convert((BC[2], CB[2]))

    ABC, BAC, CAB = \
      mpc.sign_combine((AB, AC)), \
      mpc.sign_combine((BA, BC)), \
      mpc.sign_combine((CA, CB)),

    M = b'MPC on a Friday night'

    A, B, C = \
      mpc.sign(M, (ABC[1], BAC[1], CAB[1])), \
      mpc.sign(M, (ABC[2], BAC[2], CAB[2])), \
      mpc.sign(M, (ABC[3], BAC[3], CAB[3])),


    sig = mpc.sign_combine((A, B, C))

    assert mpc.verify(M, sig)



import unittest, re, subprocess, sys
from functools import partial

def extract_results(stdout):
  r = re.compile('^([^:]+: )?([^\s]+)')
  return list(
    map(
      lambda chunk: r.match(chunk).group(2),
      filter(
        lambda line: len(line) > 0,
        stdout.decode('ascii').split('\n')
      )
    )
  )

def ggmpc(*args, **kwargs):
  process = subprocess.run(
    [sys.executable, '-m', 'src.ggmpc'] + list(args),
    capture_output=True,
  )
  if process.returncode != 0:
    raise RuntimeError(process.stderr.decode('ascii'))
  if len(process.stderr):
    sys.stderr.write(process.stderr.decode('ascii'))
  if not 'extract_results' in kwargs or kwargs['extract_results'] is True:
    return extract_results(process.stdout)

class Ecdsa(unittest.TestCase):

  ggmpc = partial(ggmpc, 'ecdsa')

  def test_2_in_2_of_3(self):
    A = Ecdsa.ggmpc('keyshare', '-i', '1', '-t', '2', '-n', '3')
    B = Ecdsa.ggmpc('keyshare', '-i', '2', '-t', '2', '-n', '3')
    C = Ecdsa.ggmpc('keyshare', '-i', '3', '-t', '2', '-n', '3')

    A, B, C = \
      Ecdsa.ggmpc('keycombine', A[0], B[1], C[1]), \
      Ecdsa.ggmpc('keycombine', A[1], B[0], C[2]), \
      Ecdsa.ggmpc('keycombine', A[2], B[2], C[0]),

    AB = Ecdsa.ggmpc('signshare', A[0], A[1])
    BA = Ecdsa.ggmpc('signconvert', B[0], B[1], AB[1])
    AB = Ecdsa.ggmpc('signconvert', AB[0], BA[1])
    BA = Ecdsa.ggmpc('signconvert', BA[0], AB[1])

    AB, BA = \
      Ecdsa.ggmpc('signcombine', AB[0]), \
      Ecdsa.ggmpc('signcombine', BA[0]),

    M = 'TOO MANY SECRETS'

    (A,), (B,) = \
      Ecdsa.ggmpc('sign', M, AB[0], BA[1]), \
      Ecdsa.ggmpc('sign', M, AB[1], BA[0]),

    sig, = Ecdsa.ggmpc('signcombine', A, B)

    Ecdsa.ggmpc('verify', M, sig, extract_results=False)

  def test_3_in_3_of_5(self):
    A = Ecdsa.ggmpc('keyshare', '-i', '1', '-t', '3', '-n', '5')
    B = Ecdsa.ggmpc('keyshare', '-i', '2', '-t', '3', '-n', '5')
    C = Ecdsa.ggmpc('keyshare', '-i', '3', '-t', '3', '-n', '5')
    D = Ecdsa.ggmpc('keyshare', '-i', '4', '-t', '3', '-n', '5')
    E = Ecdsa.ggmpc('keyshare', '-i', '5', '-t', '3', '-n', '5')

    A, B, C, D, E = \
      Ecdsa.ggmpc('keycombine', A[0], B[1], C[1], D[1], E[1]), \
      Ecdsa.ggmpc('keycombine', A[1], B[0], C[2], D[2], E[2]), \
      Ecdsa.ggmpc('keycombine', A[2], B[2], C[0], D[3], E[3]), \
      Ecdsa.ggmpc('keycombine', A[3], B[3], C[3], D[0], E[4]), \
      Ecdsa.ggmpc('keycombine', A[4], B[4], C[4], D[4], E[0]),

    A, B, C = \
      Ecdsa.ggmpc('signshare', A[0], A[1], A[2]), \
      Ecdsa.ggmpc('signshare', B[1], B[0], B[2]), \
      Ecdsa.ggmpc('signshare', C[1], C[2], C[0]),

    AB = Ecdsa.ggmpc('signconvert', A[0], B[1])
    BA = Ecdsa.ggmpc('signconvert', B[0], AB[1])
    AB = Ecdsa.ggmpc('signconvert', AB[0], BA[1])

    AC = Ecdsa.ggmpc('signconvert', A[0], C[1])
    CA = Ecdsa.ggmpc('signconvert', C[0], AC[1])
    AC = Ecdsa.ggmpc('signconvert', AC[0], CA[1])

    BC = Ecdsa.ggmpc('signconvert', B[0], C[2])
    CB = Ecdsa.ggmpc('signconvert', C[0], BC[1])
    BC = Ecdsa.ggmpc('signconvert', BC[0], CB[1])

    ABC, BAC, CAB = \
      Ecdsa.ggmpc('signcombine', AB[0], AC[0]), \
      Ecdsa.ggmpc('signcombine', BA[0], BC[0]), \
      Ecdsa.ggmpc('signcombine', CA[0], CB[0]),

    M = 'TOO MANY SECRETS'

    (A,), (B,), (C,) = \
      Ecdsa.ggmpc('sign', M, ABC[0], BAC[1], CAB[1]), \
      Ecdsa.ggmpc('sign', M, ABC[1], BAC[0], CAB[2]), \
      Ecdsa.ggmpc('sign', M, ABC[2], BAC[2], CAB[0]),

    sig, = Ecdsa.ggmpc('signcombine', A, B, C)

    Ecdsa.ggmpc('verify', M, sig, extract_results=False)

class Eddsa(unittest.TestCase):

  ggmpc = partial(ggmpc, 'eddsa')

  def test_2_in_2_of_3(self):
    A = Eddsa.ggmpc('keyshare', '-i', '1', '-t', '2', '-n', '3')
    B = Eddsa.ggmpc('keyshare', '-i', '2', '-t', '2', '-n', '3')
    C = Eddsa.ggmpc('keyshare', '-i', '3', '-t', '2', '-n', '3')

    A, B, C = \
      Eddsa.ggmpc('keycombine', A[0], B[1], C[1]), \
      Eddsa.ggmpc('keycombine', A[1], B[0], C[2]), \
      Eddsa.ggmpc('keycombine', A[2], B[2], C[0]),

    M = 'TOO MANY SECRETS'

    A, B = \
      Eddsa.ggmpc('signshare', M, A[0], A[1]), \
      Eddsa.ggmpc('signshare', M, B[0], B[1])

    (A,), (B,) = \
      Eddsa.ggmpc('sign', M, A[0], B[1]), \
      Eddsa.ggmpc('sign', M, A[1], B[0])

    sig, = Eddsa.ggmpc('signcombine', A, B)

    assert Eddsa.ggmpc('verify', M, sig)

  def test_3_in_3_of_5(self):
    A = Eddsa.ggmpc('keyshare', '-i', '1', '-t', '3', '-n', '5')
    B = Eddsa.ggmpc('keyshare', '-i', '2', '-t', '3', '-n', '5')
    C = Eddsa.ggmpc('keyshare', '-i', '3', '-t', '3', '-n', '5')
    D = Eddsa.ggmpc('keyshare', '-i', '4', '-t', '3', '-n', '5')
    E = Eddsa.ggmpc('keyshare', '-i', '5', '-t', '3', '-n', '5')

    A, B, C, D, E = \
      Eddsa.ggmpc('keycombine', A[0], B[1], C[1], D[1], E[1]), \
      Eddsa.ggmpc('keycombine', A[1], B[0], C[2], D[2], E[2]), \
      Eddsa.ggmpc('keycombine', A[2], B[2], C[0], D[3], E[3]), \
      Eddsa.ggmpc('keycombine', A[3], B[3], C[3], D[0], E[4]), \
      Eddsa.ggmpc('keycombine', A[4], B[4], C[4], D[4], E[0])

    M = 'TOO MANY SECRETS'

    A, B, C = \
      Eddsa.ggmpc('signshare', M, A[0], A[1], A[2]), \
      Eddsa.ggmpc('signshare', M, B[0], B[1], B[2]), \
      Eddsa.ggmpc('signshare', M, C[0], C[1], C[2])

    (A,), (B,), (C,) = \
      Eddsa.ggmpc('sign', M, A[0], B[1], C[1]), \
      Eddsa.ggmpc('sign', M, A[1], B[0], C[2]), \
      Eddsa.ggmpc('sign', M, A[2], B[2], C[0])

    sig, = Eddsa.ggmpc('signcombine', A, B, C)

    assert Eddsa.ggmpc('verify', M, sig)



""" Serialization functions. """

from . import ecdsa, eddsa

def serialize_int(x, length):
  """
  Serialize an integer as a big-endian byte array.

  :param x: The integer to serialize.

  :type x: int

  :param length: Byte length of serialized array.

  :type length: int

  :return: Serialized integer.

  :rtype: bytes
  """
  return x.to_bytes(length, 'big')


def deserialize_int(ser, length):
  """
  Deserialize an integer from a big-endian byte array.

  :param ser: Serialized data.

  :type ser: bytes

  :param length: Byte length of serialized integer.

  :type length: int

  :return: Deserialized integer.

  :rtype: int
  """
  x = int.from_bytes(ser[:length], 'big')
  return ser[length:], x



import ecdsa, nacl, unittest
import src.ggmpc as ggmpc
import src.ggmpc.curves as curves

class Eddsa(unittest.TestCase):

  def test_2_in_2_of_3(self):
    mpc = ggmpc.Eddsa(curves.ed25519)

    A = mpc.key_share(1, 2, 3)
    B = mpc.key_share(2, 2, 3)
    C = mpc.key_share(3, 2, 3)

    A, B, C = \
      mpc.key_combine((A[1], B[1], C[1])), \
      mpc.key_combine((A[2], B[2], C[2])), \
      mpc.key_combine((A[3], B[3], C[3]))

    M = b'MPC on a Friday night'

    A, B = mpc.sign_share(M, (A[1], A[2])), mpc.sign_share(M, (B[1], B[2]))

    A, B = mpc.sign(M, (A[1], B[1])), mpc.sign(M, (A[2], B[2]))

    sig = mpc.sign_combine((A, B))

    assert mpc.verify(M, sig)

  def test_2_in_3_of_5(self):
    mpc = ggmpc.Eddsa(curves.ed25519)

    A = mpc.key_share(1, 3, 5)
    B = mpc.key_share(2, 3, 5)
    C = mpc.key_share(3, 3, 5)
    D = mpc.key_share(4, 3, 5)
    E = mpc.key_share(5, 3, 5)

    A, B, C, D, E = \
      mpc.key_combine((A[1], B[1], C[1], D[1], E[1])), \
      mpc.key_combine((A[2], B[2], C[2], D[2], E[2])), \
      mpc.key_combine((A[3], B[3], C[3], D[3], E[3])), \
      mpc.key_combine((A[4], B[4], C[4], D[4], E[4])), \
      mpc.key_combine((A[5], B[5], C[5], D[5], E[5]))

    M = b'MPC on a Friday night'

    A, B = mpc.sign_share(M, (A[1], A[2])), mpc.sign_share(M, (B[1], B[2]))

    A, B = mpc.sign(M, (A[1], B[1])), mpc.sign(M, (A[2], B[2]))

    sig = mpc.sign_combine((A, B))

    try:
      assert not mpc.verify(M, sig)
    except nacl.exceptions.BadSignatureError:
      pass

  def test_3_in_3_of_5(self):
    mpc = ggmpc.Eddsa(curves.ed25519)

    A = mpc.key_share(1, 3, 5)
    B = mpc.key_share(2, 3, 5)
    C = mpc.key_share(3, 3, 5)
    D = mpc.key_share(4, 3, 5)
    E = mpc.key_share(5, 3, 5)

    A, B, C, D, E = \
      mpc.key_combine((A[1], B[1], C[1], D[1], E[1])), \
      mpc.key_combine((A[2], B[2], C[2], D[2], E[2])), \
      mpc.key_combine((A[3], B[3], C[3], D[3], E[3])), \
      mpc.key_combine((A[4], B[4], C[4], D[4], E[4])), \
      mpc.key_combine((A[5], B[5], C[5], D[5], E[5]))

    M = b'MPC on a Friday night'

    A, B, C = \
      mpc.sign_share(M, (A[1], A[2], A[3])), \
      mpc.sign_share(M, (B[1], B[2], B[3])), \
      mpc.sign_share(M, (C[1], C[2], C[3]))

    A, B, C = \
      mpc.sign(M, (A[1], B[1], C[1])), \
      mpc.sign(M, (A[2], B[2], C[2])), \
      mpc.sign(M, (A[3], B[3], C[3]))

    sig = mpc.sign_combine((A, B, C))

    assert mpc.verify(M, sig)



""" Serialization functions for EdDSA data. """

from .. import serialization

def serialize_u_share(share):
  """
  Serialize a u-share.

  :param share: U-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['y'], 32)
  ser += serialization.serialize_int(share['u'], 32)
  ser += serialization.serialize_int(share['prefix'], 32)
  return ser

def deserialize_u_share(ser):
  """
  Deserialize a u-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized u-share.

  :rtype: tuple
  """
  P_i = {}
  ser, P_i['i'] = serialization.deserialize_int(ser, 1)
  ser, P_i['y'] = serialization.deserialize_int(ser, 32)
  ser, P_i['u'] = serialization.deserialize_int(ser, 32)
  ser, P_i['prefix'] = serialization.deserialize_int(ser, 32)
  return ser, P_i

def serialize_y_share(share):
  """
  Serialize a y-share.

  :param share: Y-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['j'], 1)
  ser += serialization.serialize_int(share['y'], 32)
  ser += serialization.serialize_int(share['u'], 32)
  return ser

def deserialize_y_share(ser):
  """
  Deserialize a y-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized y-share.

  :rtype: tuple
  """
  P_j = {}
  ser, P_j['i'] = serialization.deserialize_int(ser, 1)
  ser, P_j['j'] = serialization.deserialize_int(ser, 1)
  ser, P_j['y'] = serialization.deserialize_int(ser, 32)
  ser, P_j['u'] = serialization.deserialize_int(ser, 32)
  return ser, P_j

def serialize_p_share(P_i):
  """
  Serialize a p-share.

  :param share: P-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(P_i['i'], 1)
  ser += serialization.serialize_int(P_i['y'], 32)
  ser += serialization.serialize_int(P_i['x'], 32)
  ser += serialization.serialize_int(P_i['prefix'], 32)
  return ser

def deserialize_p_share(ser):
  """
  Deserialize a p-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized p-share.

  :rtype: tuple
  """
  P_i = {}
  ser, P_i['i'] = serialization.deserialize_int(ser, 1)
  ser, P_i['y'] = serialization.deserialize_int(ser, 32)
  ser, P_i['x'] = serialization.deserialize_int(ser, 32)
  ser, P_i['prefix'] = serialization.deserialize_int(ser, 32)
  return ser, P_i

def serialize_j_share(P_j):
  """
  Serialize a j-share.

  :param share: J-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(P_j['i'], 1)
  ser += serialization.serialize_int(P_j['j'], 1)
  return ser

def deserialize_j_share(ser):
  """
  Deserialize a j-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized j-share.

  :rtype: tuple
  """
  P_j = {}
  ser, P_j['i'] = serialization.deserialize_int(ser, 1)
  ser, P_j['j'] = serialization.deserialize_int(ser, 1)
  return ser, P_j

def serialize_x_share(share):
  """
  Serialize an x-share.

  :param share: X-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['y'], 32)
  ser += serialization.serialize_int(share['x'], 32)
  ser += serialization.serialize_int(share['r'], 32)
  ser += serialization.serialize_int(share['R'], 32)
  return ser

def deserialize_x_share(ser):
  """
  Deserialize an x-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized x-share.

  :rtype: tuple
  """
  P_i = {}
  ser, P_i['i'] = serialization.deserialize_int(ser, 1)
  ser, P_i['y'] = serialization.deserialize_int(ser, 32)
  ser, P_i['x'] = serialization.deserialize_int(ser, 32)
  ser, P_i['r'] = serialization.deserialize_int(ser, 32)
  ser, P_i['R'] = serialization.deserialize_int(ser, 32)
  return ser, P_i

def serialize_r_share(share):
  """
  Serialize an r-share.

  :param share: R-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['j'], 1)
  ser += serialization.serialize_int(share['r'], 32)
  ser += serialization.serialize_int(share['R'], 32)
  return ser

def deserialize_r_share(ser):
  """
  Deserialize an r-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized r-share.

  :rtype: tuple
  """
  P_j = {}
  ser, P_j['i'] = serialization.deserialize_int(ser, 1)
  ser, P_j['j'] = serialization.deserialize_int(ser, 1)
  ser, P_j['r'] = serialization.deserialize_int(ser, 32)
  ser, P_j['R'] = serialization.deserialize_int(ser, 32)
  return ser, P_j

def serialize_g_share(share):
  """
  Serialize a g-share.

  :param share: G-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['y'], 32)
  ser += serialization.serialize_int(share['gamma'], 32)
  ser += serialization.serialize_int(share['R'], 32)
  return ser

def deserialize_g_share(ser):
  """
  Deserialize a g-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized g-share.

  :rtype: tuple
  """
  P_i = {}
  ser, P_i['i'] = serialization.deserialize_int(ser, 1)
  ser, P_i['y'] = serialization.deserialize_int(ser, 32)
  ser, P_i['gamma'] = serialization.deserialize_int(ser, 32)
  ser, P_i['R'] = serialization.deserialize_int(ser, 32)
  return ser, P_i

def serialize_signature(signature):
  """
  Serialize a signature.

  :param share: Signature to serialize.

  :type share: dict

  :return: Serialization of `signature`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(signature['y'], 32)
  ser += serialization.serialize_int(signature['R'], 32)
  ser += serialization.serialize_int(signature['sigma'], 32)
  return ser

def deserialize_signature(ser):
  """
  Deserialize a signature.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized 
    signature.

  :rtype: tuple
  """
  sig = {}
  ser, sig['y'] = serialization.deserialize_int(ser, 32)
  ser, sig['R'] = serialization.deserialize_int(ser, 32)
  ser, sig['sigma'] = serialization.deserialize_int(ser, 32)
  return ser, sig



""" Serialization functions for ECDSA data. """

import ecdsa

from .. import serialization

def serialize_point(p):
  """
  Serialize point.

  :param p: Point to serialize.

  :type x: ecdsa.ellipticcurve.PointJacobi

  :return: Serialization of `p`.

  :rtype: bytes
  """
  return p.to_bytes(encoding='compressed')

def deserialize_point(ser):
  """
  Deserialize a point.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized point.

  :rtype: tuple
  """
  if ser[0] == 0x00:
    return ser[1:], ecdsa.ellipticcurve.INFINITY
  if ser[0] == 0x04:
    return ser[65:], ecdsa.ellipticcurve.PointJacobi.from_bytes(
      ecdsa.ecdsa.curve_secp256k1,
      ser[:65],
    )
  return ser[33:], ecdsa.ellipticcurve.PointJacobi.from_bytes(
    ecdsa.ecdsa.curve_secp256k1,
    ser[:33],
  )

def serialize_p_share(share):
  """
  Serialize a p-share.

  :param share: P-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['p'], 192)
  ser += serialization.serialize_int(share['q'], 192)
  ser += serialize_point(share['y'])
  ser += serialization.serialize_int(share['u'], 32)
  return ser

def deserialize_p_share(ser):
  """
  Deserialize a p-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized p-share.

  :rtype: tuple
  """
  P_i = {}
  ser, P_i['i'] = serialization.deserialize_int(ser, 1)
  ser, P_i['p'] = serialization.deserialize_int(ser, 192)
  ser, P_i['q'] = serialization.deserialize_int(ser, 192)
  ser, P_i['y'] = deserialize_point(ser)
  ser, P_i['u'] = serialization.deserialize_int(ser, 32)
  return ser, P_i

def serialize_n_share(share):
  """
  Serialize an n-share.

  :param share: N-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['j'], 1)
  ser += serialization.serialize_int(share['n'], 384)
  ser += serialize_point(share['y'])
  ser += serialization.serialize_int(share['u'], 32)
  return ser

def deserialize_n_share(ser):
  """
  Deserialize an n-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized n-share.

  :rtype: tuple
  """
  P_j = {}
  ser, P_j['i'] = serialization.deserialize_int(ser, 1)
  ser, P_j['j'] = serialization.deserialize_int(ser, 1)
  ser, P_j['n'] = serialization.deserialize_int(ser, 384)
  ser, P_j['y'] = deserialize_point(ser)
  ser, P_j['u'] = serialization.deserialize_int(ser, 32)
  return ser, P_j

def serialize_x_share(share):
  """
  Serialize an x-share.

  :param share: X-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['p'], 192)
  ser += serialization.serialize_int(share['q'], 192)
  ser += serialize_point(share['y'])
  ser += serialization.serialize_int(share['x'], 32)
  return ser

def deserialize_x_share(ser):
  """
  Deserialize an x-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized x-share.

  :rtype: tuple
  """
  P_i = {}
  ser, P_i['i'] = serialization.deserialize_int(ser, 1)
  ser, P_i['p'] = serialization.deserialize_int(ser, 192)
  ser, P_i['q'] = serialization.deserialize_int(ser, 192)
  ser, P_i['y'] = deserialize_point(ser)
  ser, P_i['x'] = serialization.deserialize_int(ser, 32)
  return ser, P_i

def serialize_y_share(share):
  """
  Serialize a y-share.

  :param share: Y-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['j'], 1)
  ser += serialization.serialize_int(share['n'], 384)
  return ser

def deserialize_y_share(ser):
  """
  Deserialize a y-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized y-share.

  :rtype: tuple
  """
  P_j = {}
  ser, P_j['i'] = serialization.deserialize_int(ser, 1)
  ser, P_j['j'] = serialization.deserialize_int(ser, 1)
  ser, P_j['n'] = serialization.deserialize_int(ser, 384)
  return ser, P_j

def serialize_w_share(share):
  """
  Serialize a w-share.

  :param share: W-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['p'], 192)
  ser += serialization.serialize_int(share['q'], 192)
  ser += serialize_point(share['y'])
  ser += serialization.serialize_int(share['k'], 32)
  ser += serialization.serialize_int(share['w'], 32)
  ser += serialization.serialize_int(share['gamma'], 32)
  return ser

def deserialize_w_share(ser):
  """
  Deserialize a w-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized w-share.

  :rtype: tuple
  """
  S_i = {}
  ser, S_i['i'] = serialization.deserialize_int(ser, 1)
  ser, S_i['p'] = serialization.deserialize_int(ser, 192)
  ser, S_i['q'] = serialization.deserialize_int(ser, 192)
  ser, S_i['y'] = deserialize_point(ser)
  ser, S_i['k'] = serialization.deserialize_int(ser, 32)
  ser, S_i['w'] = serialization.deserialize_int(ser, 32)
  ser, S_i['gamma'] = serialization.deserialize_int(ser, 32)
  return ser, S_i

def serialize_k_share(share):
  """
  Serialize a k-share.

  :param share: K-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['j'], 1)
  ser += serialization.serialize_int(share['n'], 384)
  ser += serialization.serialize_int(share['k'], 768)
  return ser

def deserialize_k_share(ser):
  """
  Deserialize a k-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized k-share.

  :rtype: tuple
  """
  S_j = {}
  ser, S_j['i'] = serialization.deserialize_int(ser, 1)
  ser, S_j['j'] = serialization.deserialize_int(ser, 1)
  ser, S_j['n'] = serialization.deserialize_int(ser, 384)
  ser, S_j['k'] = serialization.deserialize_int(ser, 768)
  return ser, S_j

def serialize_b_share(shares):
  """
  Serialize a beta-share.

  :param share: Beta-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  S_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
  S_j = shares[next(filter(lambda j: 'j' in shares[j], shares))]
  ser = b''
  ser += serialization.serialize_int(S_i['i'], 1)
  ser += serialization.serialize_int(S_j['i'], 1)
  ser += serialization.serialize_int(S_i['p'], 192)
  ser += serialization.serialize_int(S_i['q'], 192)
  ser += serialize_point(S_i['y'])
  ser += serialization.serialize_int(S_i['k'], 32)
  ser += serialization.serialize_int(S_i['w'], 32)
  ser += serialization.serialize_int(S_i['gamma'], 32)
  ser += serialization.serialize_int(S_i['beta'], 32)
  ser += serialization.serialize_int(S_i['nu'], 32)
  ser += serialization.serialize_int(S_j['n'], 384)
  ser += serialization.serialize_int(S_j['k'], 768)
  ser += serialization.serialize_int(S_j['alpha'], 768)
  ser += serialization.serialize_int(S_j['mu'], 768)
  return ser

def deserialize_b_share(ser):
  """
  Deserialize a beta-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized
     beta-share.

  :rtype: tuple
  """
  S_i, S_j = {}, {}
  ser, S_i['i'] = serialization.deserialize_int(ser, 1)
  ser, S_j['i'] = serialization.deserialize_int(ser, 1)
  S_j['j'] = S_i['i']
  ser, S_i['p'] = serialization.deserialize_int(ser, 192)
  ser, S_i['q'] = serialization.deserialize_int(ser, 192)
  ser, S_i['y'] = deserialize_point(ser)
  ser, S_i['k'] = serialization.deserialize_int(ser, 32)
  ser, S_i['w'] = serialization.deserialize_int(ser, 32)
  ser, S_i['gamma'] = serialization.deserialize_int(ser, 32)
  ser, S_i['beta'] = serialization.deserialize_int(ser, 32)
  ser, S_i['nu'] = serialization.deserialize_int(ser, 32)
  ser, S_j['n'] = serialization.deserialize_int(ser, 384)
  ser, S_j['k'] = serialization.deserialize_int(ser, 768)
  ser, S_j['alpha'] = serialization.deserialize_int(ser, 768)
  ser, S_j['mu'] = serialization.deserialize_int(ser, 768)
  return ser, {
    S_i['i']: S_i,
    S_j['i']: S_j,
  }

def serialize_a_share(share):
  """
  Serialize an alpha-share.

  :param share: Alpha-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['j'], 1)
  ser += serialization.serialize_int(share['n'], 384)
  ser += serialization.serialize_int(share['k'], 768)
  ser += serialization.serialize_int(share['alpha'], 768)
  ser += serialization.serialize_int(share['mu'], 768)
  return ser

def deserialize_a_share(ser):
  """
  Deserialize an alpha-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized w-share.

  :rtype: tuple
  """
  S_j = {}
  ser, S_j['i'] = serialization.deserialize_int(ser, 1)
  ser, S_j['j'] = serialization.deserialize_int(ser, 1)
  ser, S_j['n'] = serialization.deserialize_int(ser, 384)
  ser, S_j['k'] = serialization.deserialize_int(ser, 768)
  ser, S_j['alpha'] = serialization.deserialize_int(ser, 768)
  ser, S_j['mu'] = serialization.deserialize_int(ser, 768)
  return ser, S_j

def serialize_m_share(share):
  """
  Serialize a mu-share.

  :param share: Mu-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['j'], 1)
  ser += serialization.serialize_int(share['alpha'], 768)
  ser += serialization.serialize_int(share['mu'], 768)
  return ser

def deserialize_m_share(ser):
  """
  Deserialize a mu-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized mu-share.

  :rtype: tuple
  """
  S_j = {}
  ser, S_j['i'] = serialization.deserialize_int(ser, 1)
  ser, S_j['j'] = serialization.deserialize_int(ser, 1)
  ser, S_j['alpha'] = serialization.deserialize_int(ser, 768)
  ser, S_j['mu'] = serialization.deserialize_int(ser, 768)
  return ser, S_j

def serialize_g_share(shares):
  """
  Serialize a gamma-share.

  :param share: Gamma-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  S_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
  S_j = shares[next(filter(lambda j: 'j' in shares[j], shares))]
  ser = b''
  ser += serialization.serialize_int(S_i['i'], 1)
  ser += serialization.serialize_int(S_j['i'], 1)
  ser += serialize_point(S_i['y'])
  ser += serialization.serialize_int(S_i['k'], 32)
  ser += serialization.serialize_int(S_i['w'], 32)
  ser += serialization.serialize_int(S_i['gamma'], 32)
  ser += serialization.serialize_int(S_i['alpha'], 32)
  ser += serialization.serialize_int(S_i['beta'], 32)
  ser += serialization.serialize_int(S_i['mu'], 32)
  ser += serialization.serialize_int(S_i['nu'], 32)
  return ser

def deserialize_g_share(ser):
  """
  Deserialize a gamma-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized
    gamma-share.

  :rtype: tuple
  """
  S_i, S_j = {}, {}
  ser, S_i['i'] = serialization.deserialize_int(ser, 1)
  ser, S_j['i'] = serialization.deserialize_int(ser, 1)
  S_j['j'] = S_i['i']
  ser, S_i['y'] = deserialize_point(ser)
  ser, S_i['k'] = serialization.deserialize_int(ser, 32)
  ser, S_i['w'] = serialization.deserialize_int(ser, 32)
  ser, S_i['gamma'] = serialization.deserialize_int(ser, 32)
  ser, S_i['alpha'] = serialization.deserialize_int(ser, 32)
  ser, S_i['beta'] = serialization.deserialize_int(ser, 32)
  ser, S_i['mu'] = serialization.deserialize_int(ser, 32)
  ser, S_i['nu'] = serialization.deserialize_int(ser, 32)
  return ser, {
    S_i['i']: S_i,
    S_j['i']: S_j,
  }

def serialize_o_share(share):
  """
  Serialize an omicron-share.

  :param share: Omicron-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialize_point(share['y'])
  ser += serialization.serialize_int(share['k'], 32)
  ser += serialization.serialize_int(share['omicron'], 32)
  ser += serialization.serialize_int(share['delta'], 32)
  ser += serialize_point(share['Gamma'])
  return ser

def deserialize_o_share(ser):
  """
  Deserialize a omicron-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized
    omicron-share.

  :rtype: tuple
  """
  S_i = {}
  ser, S_i['i'] = serialization.deserialize_int(ser, 1)
  ser, S_i['y'] = deserialize_point(ser)
  ser, S_i['k'] = serialization.deserialize_int(ser, 32)
  ser, S_i['omicron'] = serialization.deserialize_int(ser, 32)
  ser, S_i['delta'] = serialization.deserialize_int(ser, 32)
  ser, S_i['Gamma'] = deserialize_point(ser)
  return ser, S_i

def serialize_d_share(share):
  """
  Serialize a delta-share.

  :param share: Delta-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialization.serialize_int(share['j'], 1)
  ser += serialization.serialize_int(share['delta'], 32)
  ser += serialize_point(share['Gamma'])
  return ser

def deserialize_d_share(ser):
  """
  Deserialize a delta-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized
    delta-share.

  :rtype: tuple
  """
  S_j = {}
  ser, S_j['i'] = serialization.deserialize_int(ser, 1)
  ser, S_j['j'] = serialization.deserialize_int(ser, 1)
  ser, S_j['delta'] = serialization.deserialize_int(ser, 32)
  ser, S_j['Gamma'] = deserialize_point(ser)
  return ser, S_j

def serialize_s_share(share):
  """
  Serialize an s-share.

  :param share: S-share to serialize.

  :type share: dict

  :return: Serialization of `share`.

  :rtype: bytes
  """
  ser = b''
  ser += serialization.serialize_int(share['i'], 1)
  ser += serialize_point(share['y'])
  ser += serialization.serialize_int(share['r'], 32)
  ser += serialization.serialize_int(share['s'], 32)
  return ser

def deserialize_s_share(ser):
  """
  Deserialize a s-share.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized s-share.

  :rtype: tuple
  """
  S_i = {}
  ser, S_i['i'] = serialization.deserialize_int(ser, 1)
  ser, S_i['y'] = deserialize_point(ser)
  ser, S_i['r'] = serialization.deserialize_int(ser, 32)
  ser, S_i['s'] = serialization.deserialize_int(ser, 32)
  return ser, S_i

def serialize_signature(signature):
  """
  Serialize a signature.

  :param signature: Signature to serialize.

  :type signature: dict

  :return: Serialization of `signature`.

  :rtype: bytes
  """
  ser = b''
  ser += serialize_point(signature['y'])
  ser += serialization.serialize_int(signature['r'], 32)
  ser += serialization.serialize_int(signature['s'], 32)
  return ser

def deserialize_signature(ser):
  """
  Deserialize a signature.

  :param ser: Serialized data.

  :type ser: bytes

  :return: The remainder of the serialized data and the deserialized
     signature.

  :rtype: tuple
  """
  sig = {}
  ser, sig['y'] = deserialize_point(ser)
  ser, sig['r'] = serialization.deserialize_int(ser, 32)
  ser, sig['s'] = serialization.deserialize_int(ser, 32)
  return ser, sig



""" MPC threshold signature library. """

import phe, random
from functools import reduce
from . import shamir

class Ecdsa:
  """
  Threshold signing over the ECDSA cryptosystem.

  :param curve: The ECDSA curve to use.

  :type curve: ggmpc.Curve
  """

  def __init__(self, curve):
    self.curve = curve

  def key_share(self, i, t, n):
    """
    Generate shares for player at index `i` of key split `(t,n)` ways.

    :param i: Player index.

    :type i: int

    :param t: Signing threshold.

    :type t: int

    :param n: Number of shares.

    :type n: int

    :return: Dictionary of shares. Share at index `i` is a private p-share.
      Other indices are n-shares to be distributed to players at their
      corresponding index.

    :rtype: dict
    """
    assert i > 0 and i <= n
    # Generate additively homomorphic encryption key.
    pk, sk = phe.generate_paillier_keypair(n_length=3072)
    # Generate $u_i \in_R \Z_q$.
    u = self.curve.scalar_random()
    y = self.curve.point_mul_base(u)
    # Compute secret shares of their $u_i$.
    u = shamir.split(self.curve, u, t, n)
    P_i = {
      'i': i,
      'p': sk.p,
      'q': sk.q,
      'y': y,
      'u': u[i],
    }
    shares = {
      P_i['i']: P_i,
    }
    for i in u:
      if i != P_i['i']:
        shares[i] = {
          'i': i,
          'j': P_i['i'],
          'n': pk.n,
          'y': y,
          'u': u[i],
        }
    return shares

  def key_combine(self, P):
    """
    Combine data shared during the key generation protocol.

    :param P: Tuple of shares for every player. Must include player's private
      p-share and n-shares received from all other players.

    :type P: tuple

    :return: Dictionary of shares. Share at player's index is a private x-share.
      Other indices are y-shares to be used when generating signing shares.

    :rtype: dict
    """
    P_i = next(filter(lambda P_i: not 'j' in P_i, P))
    # Compute the public key.
    y = reduce(
      self.curve.point_add,
      map(lambda P_i: P_i['y'], P),
    )
    # Add secret shares of $x$.
    x = reduce(self.curve.scalar_add, map(lambda P_i: P_i['u'], P))
    players = {
      P_i['i']: {
        'i': P_i['i'],
        'p': P_i['p'],
        'q': P_i['q'],
        'y': y,
        'x': x,
      }
    }
    for P_j in P:
      if 'j' in P_j:
        players[P_j['j']] = {
          'i': P_i['i'],
          'j': P_j['j'],
          'n': P_j['n'],
        }
    return players

  def sign_share(self, S):
    """
    Create signing shares.

    :param S: Tuple of shares for each signer. Must include player's private
      x-share and y-shares received from all other signers.

    :type S: tuple

    :return: Dictionary of shares. Share at signer's index is a private w-share.
      Other indices are k-shares to be distributed to signers at their
      corresponding index.

    :rtype: dict
    """
    S_i = next(filter(lambda S_i: not 'j' in S_i, S))
    pk = phe.PaillierPublicKey(S_i['p'] * S_i['q'])
    # Select $k_i, \gamma_i \in_R Z_q$.
    k = self.curve.scalar_random()
    gamma = self.curve.scalar_random()
    # Map $x_i$ in $(n,t)$ to $w_i$ in $(t,t)$.
    d = reduce(
      lambda acc, S_j: self.curve.scalar_mul(
        self.curve.scalar_sub(S_j['j'], S_i['i']),
        acc,
      ),
      [S_i['i']] + list(filter(lambda S_j: 'j' in S_j, S)),
    )
    w = reduce(
      self.curve.scalar_mul,
      [S_i['x'], self.curve.scalar_invert(d)],
      reduce(
        lambda acc, S_j: self.curve.scalar_mul(S_j['j'], acc),
        [S_i['i']] + list(filter(lambda S_j: 'j' in S_j, S)),
      ),
    )
    signers = {
      S_i['i']: {
        'i': S_i['i'],
        'p': S_i['p'],
        'q': S_i['q'],
        'y': S_i['y'],
        'k': k,
        'w': w,
        'gamma': gamma,
      },
    }
    for S_j in S:
      if 'j' in S_j:
        signers[S_j['j']] = {
          'i': S_j['j'],
          'j': S_i['i'],
          'n': pk.n,
          'k': pk.encrypt(k).ciphertext(),
        }
    return signers

  def sign_convert(self, S):
    """
    Perform multiplicitive-to-additive (MtA) share conversion with another
    signer.

    :param S: Tuple of shares for this signing pair. Must include either:

      * The signer's private w-share and the k-share received from the other
        signer.

      * The signer's private x-share, the signer's y-share for the other signer,
        and the k-share received from the other signer. Use only in a 2-of-\*
        threshold setup.

      * The signer's private beta-share and the alpha-share received from the
        other signer.

      * The signer's private beta-share and the mu-share received from the other
        player.

    :type S: tuple

    :return: Dictionary of shares. Share at signer's index is either:

      * A private beta-share if `S` included a k-share.

      * A private private gamma-share if `S` included an alpha-share or a
        mu-share.

      If there is another index, it will be a share to send to the other signer.
      The share will be either:

      * An alpha-share if `S` included a k-share.

      * A mu-share if `S` included an alpha-share.

      Note that if `S` included a mu-share, there is no share to send to the
      other player and this function returns a dictionary containing a single
      index which will point to the signer's private gamma-share.

    :rtype: dict
    """
    S_i, S_j = None, None
    if len(S) > 2:
      S_j = next(filter(lambda S_j: 'k' in S_j, S))
      S = self.sign_share(list(filter(lambda S_i: not 'k' in S_i, S)))
      S_i = S[next(filter(lambda i: not 'j' in S[i], S))]
    elif len(S) == 2:
      S_i = next(filter(lambda S_i: not 'j' in S_i, S))
      S_j = next(filter(lambda S_j: 'j' in S_j, S))
    else:
      raise RuntimeError('expected at least 2 arguments')
    assert S_i['i'] == S_j['i']
    S_i = S_i.copy()
    S_j = S_j.copy()
    if 'alpha' in S_j:
      pk = phe.PaillierPublicKey(S_i['p'] * S_i['q'])
      sk = phe.PaillierPrivateKey(pk, S_i['p'], S_i['q'])
      alpha = sk.decrypt(phe.EncryptedNumber(pk, S_j['alpha']))
      S_i['alpha'] = self.curve.scalar_reduce(alpha)
      mu = sk.decrypt(phe.EncryptedNumber(pk, S_j['mu']))
      S_i['mu'] = self.curve.scalar_reduce(mu)
      del S_i['p']
      del S_i['q']
      del S_j['alpha']
      del S_j['mu']
    if 'k' in S_j:
      pk = phe.PaillierPublicKey(S_j['n'])
      k = phe.EncryptedNumber(pk, S_j['k'])
      # MtA $k_j, \gamma_i$.
      beta0 = random.SystemRandom().randint(1, pk.max_int)
      S_i['beta'] = self.curve.scalar_negate(beta0)
      alpha = S_i['gamma'] * k + pk.encrypt(beta0)
      S_j['alpha'] = alpha.ciphertext()
      # MtA $k_j, w_i$.
      nu0 = random.SystemRandom().randint(1, pk.max_int)
      S_i['nu'] = self.curve.scalar_negate(nu0)
      mu = S_i['w'] * k + pk.encrypt(nu0)
      S_j['mu'] = mu.ciphertext()
      if 'alpha' in S_i:
        del S_j['n']
        del S_j['k']
      else:
        pk = phe.PaillierPublicKey(S_i['p'] * S_i['q'])
        S_j['n'] = pk.n
        S_j['k'] = pk.encrypt(S_i['k']).ciphertext()
    if not 'alpha' in S_j and not 'k' in S_j:
      S_j = {
        'i': S_j['i'],
        'j': S_j['j'],
      }
    S_j['i'], S_j['j'] = S_j['j'], S_j['i']
    return {
      S_i['i']: S_i,
      S_j['i']: S_j,
    }

  def sign_combine(self, shares):
    """
    Combine gamma-shares for each of your signing pairs in the list of
    signers, or, combine s-shares to produce the final signature.

    :param shares: Tuple of gamma-shares or tuple of s-shares.

    :type shares: tuple

    :return: If `shares` included gamma-shares, returns a dictionary of
      shares. Share at signer's index is a private omicron-share. Other indices
      are delta-shares to distribute to all other signers. If `shares` included
      s-shares, returns a fully reconstructed signature.

    :rtype: dict
    """
    assert len(shares) > 0
    # Check if combining $s$ shares.
    if 'r' in shares[0] and 's' in shares[0]:
      assert len(shares) > 1
      # Every r must match.
      r = shares[0]['r']
      assert reduce(
        lambda a, b: a and b,
        map(lambda share: share['r'] == r, shares[1:])
      )
      s = reduce(
        self.curve.scalar_add,
        map(lambda share: share['s'], shares),
      )
      # Normalize s.
      s = self.curve.order() - s if s > self.curve.order() // 2 else s
      return {
        'y': shares[0]['y'],
        'r': r,
        's': s,
      }
    # Combine the pairs into a map of signers.
    P_i = []
    S = {}
    for pair in shares:
      if type(pair) == dict:
        for i in pair:
          if not 'j' in pair[i]:
            P_i.append(pair[i])
          else:
            S[i] = pair[i].copy()
      else:
        for share in pair:
          if not 'j' in share:
            P_i.append(share)
          else:
            S[share['i']] = share.copy()
    alpha = map(lambda P_i: P_i['alpha'], P_i)
    beta = map(lambda P_i: P_i['beta'], P_i)
    mu = map(lambda P_i: P_i['mu'], P_i)
    nu = map(lambda P_i: P_i['nu'], P_i)
    # Compute $\delta_i = k_i \gamma_i + \sum_{j \ne i} \alpha_{ij}
    #                                  + \sum_{j \ne i} \beta_{ji}$.
    delta = self.curve.scalar_add(
      self.curve.scalar_mul(P_i[0]['k'], P_i[0]['gamma']),
      self.curve.scalar_add(
        reduce(self.curve.scalar_add, alpha),
        reduce(self.curve.scalar_add, beta),
      ),
    )
    # Compute $\omicron_i = k_i \w_i + \sum_{j \ne i} \mu_{ij}
    #                                + \sum_{j \ne i} \nu_{ji}$.
    omicron = self.curve.scalar_add(
      self.curve.scalar_mul(P_i[0]['k'], P_i[0]['w']),
      self.curve.scalar_add(
        reduce(self.curve.scalar_add, mu),
        reduce(self.curve.scalar_add, nu),
      ),
    )
    Gamma = self.curve.point_mul_base(P_i[0]['gamma'])
    return reduce(
      lambda shares, i: {**shares, **{
        i: {
          'i': i,
          'j': P_i[0]['i'],
          'delta': delta,
          'Gamma': Gamma,
        },
      }},
      S,
      {
        P_i[0]['i']: {
          'i': P_i[0]['i'],
          'y': P_i[0]['y'],
          'k': P_i[0]['k'],
          'omicron': omicron,
          'delta': delta,
          'Gamma': Gamma,
        },
      }
    )

  def sign(self, M, S):
    """
    Sign a message.

    :param M: Message to sign.

    :type M: bytes

    :param S: Tuple of shares for each signer. Must include signer's private
       omicron-share and delta-shares received from other signers.

    :type shares: tuple

    :return: Signature share to be combined with all other signers' signature
      shares.

    :rtype: dict
    """
    m = int.from_bytes(self.curve.hash(M).digest(), 'big')
    S_i = next(filter(lambda share: 'k' in share, S))
    delta = reduce(
      self.curve.scalar_add,
      map(lambda S_j: S_j['delta'], S),
    )
    # Compute $R = [\Si_{i \in S} \Gamma_i]^\delta^{-1}$ and $r = H'(R)$.
    R = self.curve.point_mul(
      reduce(
        self.curve.point_add,
        map(lambda S_j: S_j['Gamma'], S),
      ),
      self.curve.scalar_invert(delta),
    )
    r = int(R.x())
    # Compute $s_i = m k_i + r \omicron_i$.
    s = self.curve.scalar_add(
      self.curve.scalar_mul(m, S_i['k']),
      self.curve.scalar_mul(r, S_i['omicron']),
    )
    return {
      'i': S_i['i'],
      'y': S_i['y'],
      'r': r,
      's': s,
    }

  def verify(self, M, sig):
    """
    Verify a signature.

    :param M: Signed message.

    :type M: bytes

    :param sig: Signature.

    :type sig: dict

    :return: True if signature is valid; False otherwise.

    :rtype: bool
    """
    return self.curve.verify(
      sig['y'],
      M,
      sig['r'].to_bytes(32, 'big') + sig['s'].to_bytes(32, 'big'),
    )

class Eddsa:
  """
  Threshold signing over the EdDSA cryptosystem.

  :param curve: The EdDSA curve to use.

  :type curve: ggmpc.Curve
  """

  def __init__(self, curve):
    self.curve = curve

  def key_share(self, i, t, n):
    """
    Generate shares for player at index `i` of key split `(t,n)` ways.

    :param i: Player index.

    :type i: int

    :param t: Signing threshold.

    :type t: int

    :param n: Number of shares.

    :type n: int

    :return: Dictionary of shares. Share at index `i` is a private u-share.
      Other indices are y-shares to be distributed to players at their
      corresponding index.

    :rtype: dict
    """
    assert i > 0 and i <= n
    sk = random.SystemRandom().randrange(2 ** 256)
    sk = sk.to_bytes((sk.bit_length() + 7) // 8, 'big')
    h = self.curve.hash(sk).digest()
    h = b'\x00' * (64 - len(h)) + h
    u = [x for x in h[0:32]]
    u[0] &= 248
    u[31] &= 63
    u[31] |= 64
    u = self.curve.scalar_reduce(int.from_bytes(bytes([0] * 32 + u), 'big'))
    y = self.curve.point_mul_base(u)
    u = shamir.split(self.curve, u, t, n)
    prefix = [x for x in h[32:]]
    prefix = [0] * 32 + prefix
    prefix = self.curve.scalar_reduce(int.from_bytes(bytes(prefix), 'big'))
    P_i = {
      'i': i,
      'y': y,
      'u': u[i],
      'prefix': prefix,
    }
    shares = {
      P_i['i']: P_i,
    }
    for i in u:
      if i != P_i['i']:
        shares[i] = {
          'i': i,
          'j': P_i['i'],
          'y': y,
          'u': u[i],
        }
    return shares

  def key_combine(self, P):
    """ 
    Combine data shared during the key generation protocol.

    :param P: Tuple of shares for every player. Must include player's private
      p-share and n-shares received from all other players.

    :type P: tuple

    :return: Dictionary of shares. Share at player's index is a private p-share.
      Other indices are j-shares to be used when generating signing shares.

    :rtype: dict
    """
    P_i = next(filter(lambda P_i: not 'j' in P_i, P))
    # Compute the public key.
    y = reduce(
      self.curve.point_add,
      map(lambda P_i: P_i['y'], P),
    )
    # Add secret shares of $x$.
    x = reduce(self.curve.scalar_add, map(lambda P_i: P_i['u'], P))
    P_i = {
      'i': P_i['i'],
      'y': y,
      'x': x,
      'prefix': P_i['prefix'],
    }
    players = {
      P_i['i']: P_i,
    }
    for P_j in P:
      if 'j' in P_j:
        players[P_j['j']] = {
          'i': P_j['j'],
          'j': P_i['i'],
        }
    return players

  def sign_share(self, M, S):
    """
    Create signing shares.

    :param M: Message to sign.

    :type M: bytes

    :param S: Tuple of shares for each signer. Must include player's private
      p-share and j-shares received from all other signers.

    :type S: tuple

    :return: Dictionary of shares. Share at signer's index is a private x-share.
      Other indices are r-shares to be distributed to signers at their
      corresponding index.

    :rtype: dict
    """
    S_i = next(filter(lambda S_i: not 'j' in S_i, S))
    I = list(map(lambda S_i: S_i['i'], S))
    digest = self.curve.hash(
      S_i['prefix'].to_bytes(32, 'big') \
      + M \
      + self.curve.scalar_random().to_bytes(32, 'big')
    ).digest()
    r = self.curve.scalar_reduce(int.from_bytes(digest, 'big'))
    R = self.curve.point_mul_base(r)
    r = shamir.split(self.curve, r, len(S), I=I)
    shares = {
      S_i['i']: {
        'i': S_i['i'],
        'y': S_i['y'],
        'x': S_i['x'],
        'r': r[S_i['i']],
        'R': R,
      },
    }
    for S_j in S:
      if 'j' in S_j:
        shares[S_j['i']] = {
          'i': S_j['i'],
          'j': S_i['i'],
          'r': r[S_j['i']],
          'R': R,
        }
    return shares

  def sign(self, M, S):
    """
    Sign a message.

    :param M: Message to sign.

    :type M: bytes

    :param S: Tuple of shares for each signer. Must include signer's private
      x-share and r-shares received from other signers.

    :type S: tuple

    :return: Signature share to be combined with all other signers' signature
      shares.

    :rtype: dict
    """
    S_i = next(filter(lambda S_i: not 'j' in S_i, S))
    R = reduce(self.curve.point_add, map(lambda S_i: S_i['R'], S))
    digest = self.curve.hash(
      R.to_bytes(32, 'little') \
      + S_i['y'].to_bytes(32, 'little') \
      + M
    ).digest()
    k = self.curve.scalar_reduce(int.from_bytes(digest, 'little'))
    r = reduce(self.curve.scalar_add, map(lambda S_i: S_i['r'], S))
    gamma = self.curve.scalar_add(r, self.curve.scalar_mul(k, S_i['x']))
    return {
      'i': S_i['i'],
      'y': S_i['y'],
      'gamma': gamma,
      'R': R,
    }

  def sign_combine(self, S):
    """
    Combine signature shares to produce the final signature.

    :param S: Tuple of gamma-shares for each signer.

    :type S: tuple

    :return: Fully reconstructed signature.

    :rtype: dict
    """
    y = next(iter(S))['y']
    R = next(iter(S))['R']
    shares = {}
    for S_i in S:
      shares[S_i['i']] = S_i['gamma']
    sigma = shamir.combine(self.curve, shares)
    return {
      'y': y,
      'R': R,
      'sigma': sigma,
    }

  def verify(self, M, sig):
    """
    Verify a signature.

    :param M: Signed message.

    :type M: bytes

    :param sig: Signature.

    :type sig: dict

    :return: True if signature is valid; False otherwise.

    :rtype: bool
    """
    return self.curve.verify(
      sig['y'].to_bytes(32, 'little'),
      M,
      sig['R'].to_bytes(32, 'little') + sig['sigma'].to_bytes(32, 'little'),
    )
